#ifdef GRIT_HLSL
#define DISCARD return
#define UNIFORM
#define SEP ,
#define LSEP
#define SEM_TEX_0 : TEXUNIT0
#define SEM_TEX_1 : TEXUNIT1
#define SEM_TEX_2 : TEXUNIT2
#define SEM_TEX_3 : TEXUNIT3
#define SEM_TEX_4 : TEXUNIT4
#define SEM_TEX_5 : TEXUNIT5
#define SEM_TEX_6 : TEXUNIT6
#define SEM_TEX_7 : TEXUNIT7
#define SEM_POSITION : POSITION
#define SEM_TANGENT : TANGENT0
#define SEM_NORMAL : NORMAL
#define SEM_COLOR : COLOR
#define SEM_COLOR1 : COLOR1
#define SEM_TEXCOORD0 : TEXCOORD0
#define SEM_TEXCOORD1 : TEXCOORD1
#define SEM_TEXCOORD2 : TEXCOORD2
#define SEM_TEXCOORD3 : TEXCOORD3
#define SEM_TEXCOORD4 : TEXCOORD4
#define SEM_TEXCOORD5 : TEXCOORD5
#define SEM_TEXCOORD6 : TEXCOORD6
#define SEM_TEXCOORD7 : TEXCOORD7
#define SEM_TEXCOORD8 : TEXCOORD8
#define SEM_TEXCOORD9 : TEXCOORD9
#define SEM_TEXCOORD10 : TEXCOORD10
#define SEM_TEXCOORD11 : TEXCOORD11
#define VERTEX_PROGRAM_DEFINITION_BEGINS void VERTEX_PROGRAM (
#define VERTEX_PROGRAM_DEFINITION_ENDS ) {
#define FRAGMENT_PROGRAM_DEFINITION_BEGINS void FRAGMENT_PROGRAM (
#define FRAGMENT_PROGRAM_DEFINITION_ENDS ) {
float4 mul4(float4x4 mat, float4 v) { return mul(mat,v); }
float3 mul3(float3x3 mat, float3 v) { return mul(mat,v); }
float saturate(float x) { return clamp(x,0.0,1.0); }
#else
#version 130
#define float4 vec4
#define float3 vec3
#define float2 vec2
#define float3x3 mat3
#define float4x4 mat4
#define DISCARD discard
#define lerp mix
#define tex2D texture
#define UNIFORM uniform
#define SEM_TEX_0
#define SEM_TEX_1
#define SEM_TEX_2
#define SEM_TEX_3
#define SEM_TEX_4
#define SEM_TEX_5
#define SEM_TEX_6
#define SEM_TEX_7
#define SEM_POSITION
#define SEM_TANGENT
#define SEM_NORMAL
float4 mul4(float4x4 mat, float4 v) { return mat * v; }
float3 mul3(float3x3 mat, float3 v) { return mat * v; }
float saturate(float x) { return clamp(x,0.0,1.0); }
#define SEM_COLOR
#define SEM_COLOR1
#define SEM_TEXCOORD0
#define SEM_TEXCOORD1
#define SEM_TEXCOORD2
#define SEM_TEXCOORD3
#define SEM_TEXCOORD4
#define SEM_TEXCOORD5
#define SEM_TEXCOORD6
#define SEM_TEXCOORD7
#define SEM_TEXCOORD8
#define SEM_TEXCOORD9
#define SEM_TEXCOORD10
#define SEP ;
#define LSEP ;
#define VERTEX_PROGRAM_DEFINITION_BEGINS
#define VERTEX_PROGRAM_DEFINITION_ENDS void VERTEX_PROGRAM(void) {
#define FRAGMENT_PROGRAM_DEFINITION_BEGINS
#define FRAGMENT_PROGRAM_DEFINITION_ENDS void FRAGMENT_PROGRAM(void) {
#endif

#define PI 3.141592653589
#define USE_BILINEAR_PCF_TAP 0

#define SHADOW_FACTOR 3000 // max distance from 0,0,0 to any point (in direction of sun)
#define PCF_SIDE 2 // PCF taps form a grid
#define FAKE_PCF_SIDE 2 // FAKE_PCF taps form a grid within each cell of the PCF grid

// PCSS params
#define SHADOW_DIST1 20
#define SHADOW_DIST2 50
#define SHADOW_DIST3 200

// fade shadow out
#define SHADOW_FADE_START 150
#define SHADOW_FADE_END 200


#define SPREAD1 4
#define SPREAD2 4
#define SPREAD3 1.1

#define SHADOW_BIAS1 SPREAD1/(-2.11+32.19*max(0.1,flat_light_exposure))
#define SHADOW_BIAS2 SPREAD2/(-.537+10.355*max(0.1,flat_light_exposure))
#define SHADOW_BIAS3 SPREAD3/(-0.187+3.533*max(flat_light_exposure,0.1))


float tap_dist (sampler2D tex, float2 suv, float our_dist)
{
        return our_dist - tex2D(tex, suv).r;
}

float tap_test (sampler2D tex, float2 suv, float our_dist)
{
        return tap_dist(tex,suv,our_dist)>0 ? 1.0 : 0.0;
}

float bilinear_tap_test (sampler2D tex, int sz, float2 suv, float d)
{
        #if USE_BILINEAR_PCF_TAP == 0
        return tap_test(tex,suv,d);
        #else
        suv *= sz;
        float2 wh = floor(suv); // whole component
        float2 fr = suv - wh; // fractional component
        return lerp(lerp(tap_test(tex,(wh+float2(0,0))/sz,d), tap_test(tex,(wh+float2(0,1))/sz,d), fr.y),
                    lerp(tap_test(tex,(wh+float2(1,0))/sz,d), tap_test(tex,(wh+float2(1,1))/sz,d), fr.y),
                    fr.x);
        #endif
}

float test_shadow (float3 vertex_ls, sampler2D tex, float our_dist,
                   float spread, float bias, float2 screen_pos)
{
        vertex_ls.xy /= vertex_ls.z;
        if (vertex_ls.x < 0) return 0;
        if (vertex_ls.y < 0) return 0;
        if (vertex_ls.x > 1) return 0;
        if (vertex_ls.y > 1) return 0;

        our_dist -= bias;
        our_dist /= SHADOW_FACTOR;


        // range (0,1) on each axis, equal border around all edges
        float2 offset = (float2(int(screen_pos.x)%FAKE_PCF_SIDE+0.5,
                                int(screen_pos.y)%FAKE_PCF_SIDE+0.5)) / FAKE_PCF_SIDE;

        if (offset.x > 0.5) offset.y = 1 - offset.y;

        float total = 0;
        for (int y=0 ; y<PCF_SIDE ; ++y) {
                for (int x=0 ; x<PCF_SIDE ; ++x) {
                        float2 p = (float2(x,y) + offset)/PCF_SIDE - float2(0.5,0.5);
                        p *= spread / 1024;
                        float2 suv = vertex_ls.xy + p;
                        total += bilinear_tap_test(tex,1024,suv,our_dist);
                }
        }

        return total / PCF_SIDE / PCF_SIDE;
}


#ifdef VERTEX_PROGRAM
UNIFORM float4x4 world;
UNIFORM float4x4 view_proj;
UNIFORM float4x4 shadow_world_view_proj1;
UNIFORM float4x4 shadow_world_view_proj2;
UNIFORM float4x4 shadow_world_view_proj3;
UNIFORM float3 camera_pos;
UNIFORM float4 light_pos_os;
UNIFORM float4 light_pos_ws;
UNIFORM float3 light_ambient;
UNIFORM float3 surf_ambient;

VERTEX_PROGRAM_DEFINITION_BEGINS
        in float3 vertex SEM_POSITION SEP
        in float3 normal SEM_NORMAL SEP
        #if USE_NORMAL_MAP==1
        in float3 tangent SEM_TANGENT SEP
        #endif
        #if USE_DIFFUSE_MAP==1 || USE_NORMAL_MAP==1 || USE_SPECULAR_MAP==1
        in float2 uv0 SEM_TEXCOORD0 SEP
        #endif
        #if USE_VERTEX_COLOURS==1
        in float4 colour SEM_COLOR SEP
        #if USE_NIGHT_COLOURS==1
        in float4 uv1 SEM_TEXCOORD1 SEP
        #endif
        #endif
        out float4 out00 SEM_TEXCOORD0 SEP
        out float4 out01 SEM_TEXCOORD1 SEP
        out float4 out02 SEM_TEXCOORD2 SEP
        out float4 out03 SEM_TEXCOORD3 SEP
        out float4 out04 SEM_TEXCOORD4 SEP
        out float4 out05 SEM_TEXCOORD5 SEP
        out float4 out06 SEM_TEXCOORD6 SEP
        out float4 out07 SEM_TEXCOORD7 SEP
        //out float4 out08 SEM_TEXCOORD8 SEP
        //out float4 out09 SEM_TEXCOORD9 SEP
        //out float4 out10 SEM_TEXCOORD10 SEP
        #ifdef GRIT_HLSL
        out float4 gl_Position : POSITION,
        out float4 gl_FrontColor : COLOR0
        #endif
VERTEX_PROGRAM_DEFINITION_ENDS
        #if USE_VERTEX_COLOURS==1
        float3 vertex_colour_ = colour.rgb;
        gl_FrontColor.a = colour.a;
        #else
        float3 vertex_colour_ = float3(0,0,0);
        gl_FrontColor.a = 1;
        #endif

        // do ambient lighting
        gl_FrontColor.rgb = light_ambient * surf_ambient;

        // lighting vectors for fragment program
        float3 vertex_to_light_ = light_pos_os.xyz;  // assume directional light
        float3 vertex_to_camera_ = normalize(camera_pos - vertex);
        float3 normal_ = normalize(normal);
        float3 tangent_ = float3(0,1,0);
        float3 binormal_ = float3(1,0,0);
        #if USE_NORMAL_MAP==1
        tangent_ = tangent;
        binormal_ = cross(normal,tangent);
        #endif

        float2 uv0_ = float2(0,0);
        #if USE_DIFFUSE_MAP==1 || USE_NORMAL_MAP==1 || USE_SPECULAR_MAP==1
        uv0_ = uv0;
        #endif

        // vertex position in light space for shadow mapping
        float4 vertex_ls1 = mul4(shadow_world_view_proj1, float4(vertex,1));
        float4 vertex_ls2 = mul4(shadow_world_view_proj2, float4(vertex,1));
        float4 vertex_ls3 = mul4(shadow_world_view_proj3, float4(vertex,1));
        float3 vertex_ls1_ = vertex_ls1.xyw;
        float3 vertex_ls2_ = vertex_ls2.xyw;
        float3 vertex_ls3_ = vertex_ls3.xyw;

        float3 vertex_ws = mul4(world, float4(vertex,1)).xyz;
        gl_Position = mul4(view_proj, float4(vertex_ws,1));
        float light_dist_ = -dot(vertex_ws,light_pos_ws.xyz);
        float cam_dist_ = gl_Position.z;

        out00 = float4(vertex_colour_, vertex_to_light_.x);
        out01 = float4(vertex_to_light_.yz, vertex_to_camera_.xy);
        out02 = float4(vertex_to_camera_.z, normal_.xyz);
        out03 = float4(tangent_.xyz, binormal_.x);
        out04 = float4(binormal_.yz, uv0_.xy);
        out05 = float4(vertex_ls1_.xyz, vertex_ls2_.x);
        out06 = float4(vertex_ls2_.yz, vertex_ls3_.xy);
        out07 = float4(vertex_ls3_.z, light_dist_, cam_dist_, 0);
}
#endif


#ifdef FRAGMENT_PROGRAM
#if USE_SPECULAR_MAP==1
        #if USE_DIFFUSE_MAP==1
                #if USE_NORMAL_MAP==1
                        UNIFORM sampler2D diffuse_map SEM_TEX_0;
                        UNIFORM sampler2D normal_map  SEM_TEX_1;
                        UNIFORM sampler2D spec_map    SEM_TEX_2;
                        UNIFORM sampler2D shadow_map1 SEM_TEX_3;
                        UNIFORM sampler2D shadow_map2 SEM_TEX_4;
                        UNIFORM sampler2D shadow_map3 SEM_TEX_5;
                #else
                        UNIFORM sampler2D diffuse_map SEM_TEX_0;
                        UNIFORM sampler2D spec_map    SEM_TEX_1;
                        UNIFORM sampler2D shadow_map1 SEM_TEX_2;
                        UNIFORM sampler2D shadow_map2 SEM_TEX_3;
                        UNIFORM sampler2D shadow_map3 SEM_TEX_4;
                #endif
        #else
                #if USE_NORMAL_MAP==1
                        UNIFORM sampler2D normal_map  SEM_TEX_0;
                        UNIFORM sampler2D spec_map    SEM_TEX_1;
                        UNIFORM sampler2D shadow_map1 SEM_TEX_2;
                        UNIFORM sampler2D shadow_map2 SEM_TEX_3;
                        UNIFORM sampler2D shadow_map3 SEM_TEX_4;
                #else
                        UNIFORM sampler2D spec_map    SEM_TEX_0;
                        UNIFORM sampler2D shadow_map1 SEM_TEX_1;
                        UNIFORM sampler2D shadow_map2 SEM_TEX_2;
                        UNIFORM sampler2D shadow_map3 SEM_TEX_3;
                #endif
        #endif
#else
        #if USE_DIFFUSE_MAP==1
                #if USE_NORMAL_MAP==1
                        UNIFORM sampler2D diffuse_map SEM_TEX_0;
                        UNIFORM sampler2D normal_map  SEM_TEX_1;
                        UNIFORM sampler2D shadow_map1 SEM_TEX_2;
                        UNIFORM sampler2D shadow_map2 SEM_TEX_3;
                        UNIFORM sampler2D shadow_map3 SEM_TEX_4;
                #else
                        UNIFORM sampler2D diffuse_map SEM_TEX_0;
                        UNIFORM sampler2D shadow_map1 SEM_TEX_1;
                        UNIFORM sampler2D shadow_map2 SEM_TEX_2;
                        UNIFORM sampler2D shadow_map3 SEM_TEX_3;
                #endif
        #else
                #if USE_NORMAL_MAP==1
                        UNIFORM sampler2D normal_map  SEM_TEX_0;
                        UNIFORM sampler2D shadow_map1 SEM_TEX_1;
                        UNIFORM sampler2D shadow_map2 SEM_TEX_2;
                        UNIFORM sampler2D shadow_map3 SEM_TEX_3;
                #else
                        UNIFORM sampler2D shadow_map1 SEM_TEX_0;
                        UNIFORM sampler2D shadow_map2 SEM_TEX_1;
                        UNIFORM sampler2D shadow_map3 SEM_TEX_2;
                #endif
        #endif
#endif
UNIFORM float4 the_fog_params;
UNIFORM float3 the_fog_colour;
UNIFORM float4 visibility;
UNIFORM float alpha_rej;
UNIFORM float3 light_diffuse;
UNIFORM float3 light_specular;
UNIFORM float3 surf_diffuse;
UNIFORM float3 surf_specular;
UNIFORM float surf_shininess;

FRAGMENT_PROGRAM_DEFINITION_BEGINS
        #ifdef GRIT_HLSL
        in float4 gl_Color : COLOR0,
        #endif
        in float4 out00 SEM_TEXCOORD0 SEP
        in float4 out01 SEM_TEXCOORD1 SEP
        in float4 out02 SEM_TEXCOORD2 SEP
        in float4 out03 SEM_TEXCOORD3 SEP
        in float4 out04 SEM_TEXCOORD4 SEP
        in float4 out05 SEM_TEXCOORD5 SEP
        in float4 out06 SEM_TEXCOORD6 SEP
        in float4 out07 SEM_TEXCOORD7 SEP
        //in float4 out08 SEM_TEXCOORD8 SEP
        //in float4 out09 SEM_TEXCOORD9 SEP
        //in float4 out10 SEM_TEXCOORD10 SEP
        #ifdef GRIT_HLSL
        in float2 gl_FragCoord : VPOS,
        #endif
        out float4 pixel SEM_COLOR LSEP
FRAGMENT_PROGRAM_DEFINITION_ENDS
        float3 vertex_colour_ = out00.xyz;
        float3 vertex_to_light_ = float3(out00.w, out01.xy);
        float3 vertex_to_camera_ = float3(out01.zw, out02.x);
        float3 normal_ = out02.yzw;
        float3 tangent_ = out03.xyz;
        float3 binormal_ = float3(out03.w, out04.xy);
        float2 uv0_ = out04.zw;
        float3 vertex_ls1_ = out05.xyz;
        float3 vertex_ls2_ = float3(out05.w, out06.xy);
        float3 vertex_ls3_ = float3(out06.zw, out07.x);
        float light_dist_ = out07.y;
        float cam_dist_ = out07.z;

        float2 screen_pos = gl_FragCoord.xy;

        // start with vertex lighting, if any
        float3 dyn_light = vertex_colour_;
        float3 spec_light = float3(0,0,0);

        float3 normal = normal_;
        float flat_light_exposure = max(0,dot(vertex_to_light_,normal));
        float3 nm_sample = float3(0,0,0);
        #if USE_NORMAL_MAP==1
        {
                float3x3 tform = float3x3(tangent_, binormal_, normal_); 
                #ifdef GRIT_HLSL
                tform = transpose(tform);
                #endif
                nm_sample = float3(-2,2,2)*(tex2D(normal_map, uv0_).xyz - .5);
                normal = mul3(tform, nm_sample);
        }
        #endif

        // compute diffuse lighting
        float light_exposure = max(0,dot(vertex_to_light_,normal));
        dyn_light += light_exposure * light_diffuse * surf_diffuse;

        // compute specular lighting
        float3 r = reflect(-vertex_to_light_,normal);
        spec_light = pow(max(0.00001,dot(r,vertex_to_camera_)),surf_shininess)
                     * light_specular * surf_specular;
        #if USE_SPECULAR_MAP==1
        spec_light *= tex2D(spec_map, uv0_).xyz;
        #endif


        // compute shadow (reduces dynamic lighting if in shadow)
        float shadowyness = 0;
        if (cam_dist_ < SHADOW_DIST1) {
                shadowyness = test_shadow(vertex_ls1_, shadow_map1, light_dist_, SPREAD1, SHADOW_BIAS1, screen_pos);
        } else if (cam_dist_ < SHADOW_DIST2) {
                shadowyness = test_shadow(vertex_ls2_, shadow_map2, light_dist_, SPREAD2, SHADOW_BIAS2, screen_pos);
        } else if (cam_dist_ < SHADOW_DIST3) {
                shadowyness = test_shadow(vertex_ls3_, shadow_map3, light_dist_, SPREAD3, SHADOW_BIAS3, screen_pos);
        }

        float fade = min(1, (SHADOW_FADE_END-cam_dist_)/(SHADOW_FADE_END-SHADOW_FADE_START));
        shadowyness *= fade;

        // shadow affects dynamic and specular light
        dyn_light.rgb -= dyn_light.rgb * shadowyness;
        spec_light.rgb -= spec_light.rgb * shadowyness;


        // ambient as pased from vertex shader, with the dyn_light just computed
        // note that alpha in vertex colours comes from gl_Color, not dyn_light_
        pixel.rgb = gl_Color.rgb + dyn_light;
        pixel.a = gl_Color.a;

        #if USE_DIFFUSE_MAP==1
        float4 texel = tex2D(diffuse_map, uv0_);
        //pixel *= float4((normal_ + float3(1,1,1))/2,1);
        pixel *= texel;
        #endif

        pixel.rgb += spec_light;


        // cull pixels after lighting and texture, but before fading
        if (pixel.a <= alpha_rej) DISCARD;

        // exp2 fog
        float clarity = the_fog_params.x * cam_dist_;
        clarity = clamp(exp(-clarity * clarity),0.0,1.0);
        pixel.rgb = lerp(the_fog_colour, pixel.rgb, clarity);

        pixel.a *= visibility.x;
}
#endif

// vim: ts=8:sw=8:et
